from typing import Callable, Optional
from websockets.sync.client import connect
from lmnr.types import DeregisterDebuggerRequest, NodeInput, RegisterDebuggerRequest, SDKError, ToolCall
import uuid
import json
from threading import Thread

class RemoteDebugger:
    def __init__(self, project_api_key: str, tools: list[Callable[..., NodeInput]] = []):
        self.project_api_key = project_api_key
        self.url = 'wss://api.lmnr.ai/v2/endpoint/ws'
        self.tools = tools
        self.thread = Thread(target=self._run)
        self.stop_flag = False
        self.session = None
    
    def start(self) -> Optional[str]:
        self.stop_flag = False
        self.session = self._generate_session_id()
        self.thread.start()
        return self.session
    
    def stop(self):
        self.stop_flag = True
        self.thread.join()
        self.session = None
        # python allows running threads only once, so we need to create a new thread
        # in case the user wants to start the debugger again
        self.thread = Thread(target=self._run)
    
    def get_session_id(self) -> str:
        return self.session
    
    def _run(self):
        request = RegisterDebuggerRequest(debuggerSessionId=self.session)
        with connect(
            self.url,
            additional_headers={
                'Authorization': f'Bearer {self.project_api_key}'
            }
        ) as websocket:
            websocket.send(request.model_dump_json())
            print(self._format_session_id())
            while not self.stop_flag:
                try:
                    # blocks the thread until a message is received or a timeout (3 seconds) occurs
                    message = websocket.recv(3)
                except TimeoutError:
                    continue
                try:
                    tool_call = ToolCall.model_validate_json(message)
                except:
                    raise SDKError(f'Invalid message received:\n{message}')
                matching_tools = [
                    tool for tool in self.tools
                    if tool.__name__ == tool_call.function.name
                ]
                if not matching_tools:
                    raise SDKError(
                        f'Tool {tool_call.function.name} not found.'
                        ' Registered tools: '
                        f'{", ".join([tool.__name__ for tool in self.tools])}'
                    )
                tool = matching_tools[0]
                if tool.__name__ == tool_call.function.name:
                    # default the arguments to an empty dictionary
                    arguments = {}
                    try:
                        arguments = json.loads(tool_call.function.arguments)
                    except:
                        pass
                    response = tool(**arguments)
                    websocket.send(json.dumps(response))
            websocket.send(DeregisterDebuggerRequest(debuggerSessionId=self.session, deregister=True).model_dump_json())

    def _generate_session_id(self) -> str:
        return uuid.uuid4().urn[9:]

    def _format_session_id(self) -> str:
        return \
f"""
========================================
Debugger Session ID:
{self.session}
========================================
"""
    


from .sdk.endpoint import Laminar
from .types import (
    ChatMessage,
    EndpointRunError,
    EndpointRunResponse,
    NodeInput
)
from .sdk.remote_debugger import RemoteDebugger as LaminarRemoteDebugger


import threading
from dataclasses import dataclass
from typing import Union

from lmnr_engine.types import Message


@dataclass
class State:
    status: str  # "Success", "Empty", "Termination"  # TODO: Turn into Enum
    message: Union[Message, None]

    @classmethod
    def new(cls, val: Message) -> "State":
        return cls(
            status="Success",
            message=val,
        )

    @classmethod
    def empty(cls) -> "State":
        return cls(
            status="Empty",
            message=Message.empty(),
        )

    @classmethod
    def termination(cls) -> "State":
        return cls(
            status="Termination",
            message=None,
        )

    def is_success(self) -> bool:
        return self.status == "Success"

    def is_termination(self) -> bool:
        return self.status == "Termination"

    def get_out(self) -> Message:
        if self.message is None:
            raise ValueError("Cannot get message from a termination state")

        return self.message


@dataclass
class ExecState:
    output: State
    semaphore: threading.Semaphore

    @classmethod
    def new(cls) -> "ExecState":
        return cls(
            output=State.empty(),
            semaphore=threading.Semaphore(0),
        )

    # Assume this is called by the caller who doesn't need to acquire semaphore
    def set_state(self, output: State):
        self.output = output

    # Assume the caller is smart to call this after acquiring the semaphore
    def get_state(self) -> State:
        return self.output

    def set_state_and_permits(self, output: State, permits: int):
        self.output = output
        self.semaphore.release(permits)



from typing import Callable, Union

from .action import RunOutput
from .state import ExecState
from lmnr_engine.types import NodeInput


class Task:
    # unique identifier
    name: str
    # mapping from current node's handle name to previous node's unique name
    # assumes nodes have only one output
    handles_mapping: list[tuple[str, str]]
    # Value or a function that returns a value
    # Usually a function which waits for inputs from previous nodes
    value: Union[NodeInput, Callable[..., RunOutput]]  # TODO: Type this fully
    # unique node names of previous nodes
    prev: list[str]
    # unique node names of next nodes
    next: list[str]
    input_states: dict[str, ExecState]

    def __init__(
        self,
        name: str,
        handles_mapping: list[tuple[str, str]],
        value: Union[NodeInput, Callable[..., RunOutput]],
        prev: list[str],
        next: list[str],
    ) -> None:
        self.name = name
        self.handles_mapping = handles_mapping
        self.value = value
        self.prev = prev
        self.next = next
        self.input_states = {
            handle_name: ExecState.new() for (handle_name, _) in self.handles_mapping
        }



import requests
import json

from lmnr_engine.engine.action import NodeRunError, RunOutput
from lmnr_engine.types import ChatMessage, NodeInput


{% for task in cookiecutter._tasks.values() %}
{% if task.node_type == "LLM" %}
def {{task.function_name}}({{ task.handle_args }}, _env: dict[str, str]) -> RunOutput:
    {% set chat_messages_found = false %}
    {% for input_handle_name in task.input_handle_names %}
    {% if input_handle_name == 'chat_messages' %}
    {% set chat_messages_found = true %}
    {% endif %}
    {% endfor %}

    {% if chat_messages_found %}
    input_chat_messages = chat_messages
    {% else %}
    input_chat_messages = []
    {% endif %}

    rendered_prompt = """{{task.config.prompt}}"""
    {% set prompt_variables = task.input_handle_names|reject("equalto", "chat_messages") %}
    {% for prompt_variable in prompt_variables %}
    # TODO: Fix this. Using double curly braces in quotes because normal double curly braces
    # get replaced during rendering by Cookiecutter. This is a hacky solution.
    rendered_prompt = rendered_prompt.replace("{{'{{'}}{{prompt_variable}}{{'}}'}}", {{prompt_variable}})  # type: ignore
    {% endfor %}

    {% if task.config.model_params == none %}
    params = {}
    {% else %}
    params = json.loads(
        """{{task.config.model_params}}"""
    )
    {% endif %}

    messages = [ChatMessage(role="system", content=rendered_prompt)]
    messages.extend(input_chat_messages)

    {% if task.config.provider == "openai" %}
    message_jsons = [
        {"role": message.role, "content": message.content} for message in messages
    ]

    data = {
        "model": "{{task.config.model}}",
        "messages": message_jsons,
    }
    data.update(params)

    headers = {
        "Content-Type": "application/json",
        "Authorization": f"Bearer {_env['OPENAI_API_KEY']}",
    }
    res = requests.post(
        "https://api.openai.com/v1/chat/completions", json=data, headers=headers
    )

    if res.status_code != 200:
        res_json = res.json()
        raise NodeRunError(f'OpenAI completions request failed: {res_json["error"]["message"]}')

    chat_completion = res.json()

    completion_message = chat_completion["choices"][0]["message"]["content"]

    meta_log = {}
    meta_log["node_chunk_id"] = None  # TODO: Add node chunk id
    meta_log["model"] = "{{task.config.model}}"
    meta_log["prompt"] = rendered_prompt
    meta_log["input_message_count"] = len(messages)
    meta_log["input_token_count"] = chat_completion["usage"]["prompt_tokens"]
    meta_log["output_token_count"] = chat_completion["usage"]["completion_tokens"]
    meta_log["total_token_count"] = (
        chat_completion["usage"]["prompt_tokens"] + chat_completion["usage"]["completion_tokens"]
    )
    meta_log["approximate_cost"] = None  # TODO: Add approximate cost
    {% elif task.config.provider == "anthropic" %}
    data = {
        "model": "{{task.config.model}}",
        "max_tokens": 4096,
    }
    data.update(params)

    # TODO: Generate appropriate code based on this if-else block
    if len(messages) == 1 and messages[0].role == "system":
        messages[0].role = "user"
        message_jsons = [
            {"role": message.role, "content": message.content} for message in messages
        ]
        data["messages"] = message_jsons
    else:
        data["system"] = messages[0].content
        message_jsons = [
            {"role": message.role, "content": message.content} for message in messages[1:]
        ]
        data["messages"] = message_jsons

    headers = {
        "Content-Type": "application/json",
        "X-Api-Key": _env['ANTHROPIC_API_KEY'],
        "Anthropic-Version": "2023-06-01",
    }
    res = requests.post(
        "https://api.anthropic.com/v1/messages", json=data, headers=headers
    )

    if res.status_code != 200:
        raise NodeRunError(f"Anthropic message request failed: {res.text}")

    chat_completion = res.json()

    completion_message = chat_completion["content"][0]["text"]

    meta_log = {}
    meta_log["node_chunk_id"] = None  # TODO: Add node chunk id
    meta_log["model"] = "{{task.config.model}}"
    meta_log["prompt"] = rendered_prompt
    meta_log["input_message_count"] = len(messages)
    meta_log["input_token_count"] = chat_completion["usage"]["input_tokens"]
    meta_log["output_token_count"] = chat_completion["usage"]["output_tokens"]
    meta_log["total_token_count"] = (
        chat_completion["usage"]["input_tokens"] + chat_completion["usage"]["output_tokens"]
    )
    meta_log["approximate_cost"] = None  # TODO: Add approximate cost
    {% else %}
    {% endif %}

    return RunOutput(status="Success", output=completion_message)


{% elif task.node_type == "Output" %}
def {{task.function_name}}(output: NodeInput, _env: dict[str, str]) -> RunOutput:
    return RunOutput(status="Success", output=output)


{% elif task.node_type == "Input" %}
{# Do nothing for Input tasks #}
{% else %}
def {{task.function_name}}(output: NodeInput, _env: dict[str, str]) -> RunOutput:
    return RunOutput(status="Success", output=output)


{% endif %}
{% endfor %}
# Other functions can be added here



from dataclasses import dataclass
from typing import Union

from lmnr_engine.types import NodeInput


@dataclass
class RunOutput:
    status: str  # "Success" | "Termination"  TODO: Turn into Enum
    output: Union[NodeInput, None]


class NodeRunError(Exception):
    pass



import requests
from dotenv import load_dotenv
import os
import click
import logging
from cookiecutter.main import cookiecutter
from importlib import resources as importlib_resources
from pydantic.alias_generators import to_pascal

from .parser.parser import runnable_graph_to_template_vars
import lmnr

logger = logging.getLogger(__name__)


@click.group()
@click.version_option()
def cli():
    "CLI for Laminar AI Engine"


@cli.command(name="pull")
@click.argument("pipeline_name")
@click.argument("pipeline_version_name")
@click.option(
    "-p",
    "--project-api-key",
    help="Project API key",
)
@click.option(
    "-l",
    "--loglevel",
    help="Sets logging level",
)
def pull(pipeline_name, pipeline_version_name, project_api_key, loglevel):
    loglevel_str_to_val = {
        "DEBUG": logging.DEBUG,
        "INFO": logging.INFO,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
    }
    logging.basicConfig()
    logging.getLogger().setLevel(loglevel_str_to_val.get(loglevel, logging.WARNING))

    project_api_key = project_api_key or os.environ.get("LMNR_PROJECT_API_KEY")
    if not project_api_key:
        load_dotenv()
        project_api_key = os.environ.get("LMNR_PROJECT_API_KEY")
    if not project_api_key:
        raise ValueError("LMNR_PROJECT_API_KEY is not set")

    headers = {"Authorization": f"Bearer {project_api_key}"}
    params = {
        "pipelineName": pipeline_name,
        "pipelineVersionName": pipeline_version_name,
    }
    res = requests.get(
        "https://api.lmnr.ai/v2/pipeline-version-by-name",
        headers=headers,
        params=params,
    )
    if res.status_code != 200:
        try:
            res_json = res.json()
        except Exception:
            raise ValueError(
                f"Error in fetching pipeline version: {res.status_code}\n{res.text}"
            )
        raise ValueError(
            f"Error in fetching pipeline version: {res.status_code}\n{res_json}"
        )

    pipeline_version = res.json()

    class_name = to_pascal(pipeline_name.replace(" ", "_").replace("-", "_"))

    context = {
        "pipeline_name": pipeline_name,
        "pipeline_version_name": pipeline_version_name,
        "class_name": class_name,
        # _tasks starts from underscore because we don't want it to be templated
        # some tasks contains LLM nodes which have prompts
        # which we don't want to be rendered by cookiecutter
        "_tasks": runnable_graph_to_template_vars(pipeline_version["runnableGraph"]),
    }

    logger.info(f"Context:\n{context}")
    cookiecutter(
        str(importlib_resources.files(lmnr)),
        output_dir=".",
        config_file=None,
        extra_context=context,
        directory="cli/",
        no_input=True,
        overwrite_if_exists=True,
    )



from dataclasses import dataclass
from typing import Union
import uuid
import datetime
from lmnr.types import NodeInput, ChatMessage


@dataclass
class Message:
    id: uuid.UUID
    # output value of producing node in form of NodeInput
    # for the following consumer
    value: NodeInput
    # all input messages to this node; accumulates previous messages too
    # input_messages: list["Message"]
    start_time: datetime.datetime
    end_time: datetime.datetime
    # node_id: uuid.UUID
    # node_name: str
    # node_type: str
    # all node per-run metadata that needs to be logged at the end of execution
    # meta_log: MetaLog | None

    @classmethod
    def empty(cls) -> "Message":
        return cls(
            id=uuid.uuid4(),
            value="",
            # input_messages=[],
            start_time=datetime.datetime.now(),
            end_time=datetime.datetime.now(),
            # node_id=uuid.uuid4(),
            # node_name="",
            # node_type="",
        )



# Convert a list of handles to a map of input handle names
# to their respective values
import uuid
from .nodes import Handle


def map_handles(
    inputs: list[Handle],
    inputs_mappings: dict[uuid.UUID, uuid.UUID],
    output_handle_id_to_node_name: dict[str, str],
) -> list[tuple[str, str]]:
    mapping = []

    for to, from_ in inputs_mappings.items():
        for input in inputs:
            if input.id == to:
                mapping.append((input.name, from_))
                break
        else:
            raise ValueError(f"Input handle {to} not found in inputs")

    return [
        (input_name, output_handle_id_to_node_name[str(output_id)])
        for input_name, output_id in mapping
    ]



from dataclasses import dataclass
from typing import Any, Optional, Union
import uuid
from lmnr.cli.parser.nodes import Handle, HandleType, NodeFunctions
from lmnr.cli.parser.utils import map_handles
from lmnr.types import NodeInput, ChatMessage


def node_input_from_json(json_val: Any) -> NodeInput:
    if isinstance(json_val, str):
        return json_val
    elif isinstance(json_val, list):
        return [ChatMessage.model_validate(msg) for msg in json_val]
    else:
        raise ValueError(f"Invalid NodeInput value: {json_val}")


# TODO: Convert to Pydantic
@dataclass
class InputNode(NodeFunctions):
    id: uuid.UUID
    name: str
    outputs: list[Handle]
    input: Optional[NodeInput]
    input_type: HandleType

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        return []

    def node_type(self) -> str:
        return "Input"

    def config(self) -> dict:
        return {}


# TODO: Convert to Pydantic
@dataclass
class LLMNode(NodeFunctions):
    id: uuid.UUID
    name: str
    inputs: list[Handle]
    dynamic_inputs: list[Handle]
    outputs: list[Handle]
    inputs_mappings: dict[uuid.UUID, uuid.UUID]
    prompt: str
    model: str
    model_params: Optional[str]
    stream: bool
    structured_output_enabled: bool
    structured_output_max_retries: int
    structured_output_schema: Optional[str]
    structured_output_schema_target: Optional[str]

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        combined_inputs = self.inputs + self.dynamic_inputs
        return map_handles(
            combined_inputs, self.inputs_mappings, output_handle_id_to_node_name
        )

    def node_type(self) -> str:
        return "LLM"

    def config(self) -> dict:
        # For easier access in the template separate the provider and model here
        provider, model = self.model.split(":", maxsplit=1)

        return {
            "prompt": self.prompt,
            "provider": provider,
            "model": model,
            "model_params": self.model_params,
            "stream": self.stream,
            "structured_output_enabled": self.structured_output_enabled,
            "structured_output_max_retries": self.structured_output_max_retries,
            "structured_output_schema": self.structured_output_schema,
            "structured_output_schema_target": self.structured_output_schema_target,
        }


# TODO: Convert to Pydantic
@dataclass
class OutputNode(NodeFunctions):
    id: uuid.UUID
    name: str
    inputs: list[Handle]
    outputs: list[Handle]
    inputs_mappings: dict[uuid.UUID, uuid.UUID]

    def handles_mapping(
        self, output_handle_id_to_node_name: dict[str, str]
    ) -> list[tuple[str, str]]:
        return map_handles(
            self.inputs, self.inputs_mappings, output_handle_id_to_node_name
        )

    def node_type(self) -> str:
        return "Output"

    def config(self) -> dict:
        return {}


Node = Union[InputNode, OutputNode, LLMNode]


def node_from_dict(node_dict: dict) -> Node:
    if node_dict["type"] == "Input":
        return InputNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            input=node_input_from_json(node_dict["input"]),
            input_type=node_dict["inputType"],
        )
    elif node_dict["type"] == "Output":
        return OutputNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
        )
    elif node_dict["type"] == "LLM":
        return LLMNode(
            id=uuid.UUID(node_dict["id"]),
            name=node_dict["name"],
            inputs=[Handle.from_dict(handle) for handle in node_dict["inputs"]],
            dynamic_inputs=[
                Handle.from_dict(handle) for handle in node_dict["dynamicInputs"]
            ],
            outputs=[Handle.from_dict(handle) for handle in node_dict["outputs"]],
            inputs_mappings={
                uuid.UUID(k): uuid.UUID(v)
                for k, v in node_dict["inputsMappings"].items()
            },
            prompt=node_dict["prompt"],
            model=node_dict["model"],
            model_params=(
                node_dict["modelParams"] if "modelParams" in node_dict else None
            ),
            stream=False,
            # TODO: Implement structured output
            structured_output_enabled=False,
            structured_output_max_retries=3,
            structured_output_schema=None,
            structured_output_schema_target=None,
        )
    else:
        raise ValueError(f"Node type {node_dict['type']} not supported")




import requests
import pydantic
from typing import Union, Optional

class ChatMessage(pydantic.BaseModel):
    role: str
    content: str

NodeInput = Union[str, list[ChatMessage]] # TypeAlias

class EndpointRunRequest(pydantic.BaseModel):
    inputs: dict[str, NodeInput]
    endpoint: str
    env: dict[str, str] = pydantic.Field(default_factory=dict)
    metadata: dict[str, str] = pydantic.Field(default_factory=dict)

class EndpointRunResponse(pydantic.BaseModel):
    outputs: dict[str, dict[str, NodeInput]]
    run_id: str

class EndpointRunError(Exception):
    error_code: str
    error_message: str

    def __init__(self, response: requests.Response):
        try:
            resp_json = response.json()
            self.error_code = resp_json['error_code']
            self.error_message = resp_json['error_message']
            super().__init__(self.error_message)
        except:
            super().__init__(response.text)
    
    def __str__(self) -> str:
        try:
            return str({'error_code': self.error_code, 'error_message': self.error_message})
        except:
            return super().__str__()
        
class SDKError(Exception):
    def __init__(self, error_mesasge: str):
        super().__init__(error_mesasge)

class ToolCallRequest(pydantic.BaseModel):
    name: str
    arguments: str

class ToolCall(pydantic.BaseModel):
    id: Optional[str]
    type: Optional[str]
    function: ToolCallRequest

# TODO: allow snake_case and manually convert to camelCase
class RegisterDebuggerRequest(pydantic.BaseModel):
    debuggerSessionId: str

class DeregisterDebuggerRequest(pydantic.BaseModel):
    debuggerSessionId: str
    deregister: bool



from concurrent.futures import ThreadPoolExecutor
import datetime
import logging
from typing import Optional
import uuid
from dataclasses import dataclass
import queue

from .task import Task
from .action import NodeRunError, RunOutput
from .state import State
from lmnr_engine.types import Message, NodeInput


logger = logging.getLogger(__name__)


@dataclass
class ScheduledTask:
    status: str  # "Task" | "Err"  TODO: Use an enum
    task_name: Optional[str]


class RunError(Exception):
    outputs: dict[str, Message]


@dataclass
class Engine:
    tasks: dict[str, Task]
    active_tasks: set[str]
    depths: dict[str, int]
    outputs: dict[str, Message]
    env: dict[str, str]
    thread_pool_executor: ThreadPoolExecutor
    # TODO: Store thread pool executor's Futures here to have control
    # over them (e.g. cancel them)

    @classmethod
    def new(
        cls, thread_pool_executor: ThreadPoolExecutor, env: dict[str, str] = {}
    ) -> "Engine":
        return cls(
            tasks={},
            active_tasks=set(),
            depths={},
            outputs={},
            env=env,
            thread_pool_executor=thread_pool_executor,
        )

    @classmethod
    def with_tasks(
        cls,
        tasks: list[Task],
        thread_pool_executor: ThreadPoolExecutor,
        env: dict[str, str] = {},
    ) -> "Engine":
        dag = cls.new(thread_pool_executor, env=env)

        for task in tasks:
            dag.tasks[task.name] = task
            dag.depths[task.name] = 0

        return dag

    def override_inputs(self, inputs: dict[str, NodeInput]) -> None:
        for task in self.tasks.values():
            # TODO: Check that it's the Input type task
            if not task.prev:
                task.value = inputs[task.name]

    def run(self, inputs: dict[str, NodeInput]) -> dict[str, Message]:
        self.override_inputs(inputs)

        q = queue.Queue()

        input_tasks = []
        for task in self.tasks.values():
            if len(task.prev) == 0:
                input_tasks.append(task.name)

        for task_id in input_tasks:
            q.put(ScheduledTask(status="Task", task_name=task_id))

        while True:
            logger.info("Waiting for task from queue")
            scheduled_task: ScheduledTask = q.get()
            logger.info(f"Got task from queue: {scheduled_task}")
            if scheduled_task.status == "Err":
                # TODO: Abort all other threads
                raise RunError(self.outputs)

            task: Task = self.tasks[scheduled_task.task_name]  # type: ignore
            logger.info(f"Task next: {task.next}")

            if not task.next:
                try:
                    fut = self.execute_task(task, q)
                    fut.result()
                    if not self.active_tasks:
                        return self.outputs
                except Exception:
                    raise RunError(self.outputs)
            else:
                self.execute_task(task, q)

    def execute_task_inner(
        self,
        task: Task,
        queue: queue.Queue,
    ) -> None:
        task_id = task.name
        next = task.next
        input_states = task.input_states
        active_tasks = self.active_tasks
        tasks = self.tasks
        depths = self.depths
        depth = depths[task.name]
        outputs = self.outputs

        inputs: dict[str, NodeInput] = {}
        input_messages = []

        # Wait for inputs for this task to be set
        for handle_name, input_state in input_states.items():
            logger.info(f"Task {task_id} waiting for semaphore for {handle_name}")
            input_state.semaphore.acquire()
            logger.info(f"Task {task_id} acquired semaphore for {handle_name}")

            # Set the outputs of predecessors as inputs of the current
            output = input_state.get_state()
            # If at least one of the inputs is termination,
            # also terminate this task early and set its state to termination
            if output.status == "Termination":
                return
            message = output.get_out()

            inputs[handle_name] = message.value
            input_messages.append(message)

        start_time = datetime.datetime.now()

        try:
            if callable(task.value):
                res = task.value(**inputs, _env=self.env)
            else:
                res = RunOutput(status="Success", output=task.value)

            if res.status == "Success":
                id = uuid.uuid4()
                state = State.new(
                    Message(
                        id=id,
                        value=res.output,  # type: ignore
                        start_time=start_time,
                        end_time=datetime.datetime.now(),
                    )
                )
            else:
                assert res.status == "Termination"
                state = State.termination()

            is_termination = state.is_termination()
            logger.info(f"Task {task_id} executed")

            # remove the task from active tasks once it's done
            if task_id in active_tasks:
                active_tasks.remove(task_id)

            if depth > 0:
                # propagate reset once we enter the loop
                # TODO: Implement this for cycles
                raise NotImplementedError()

            if depth == 10:
                # TODO: Implement this for cycles
                raise NotImplementedError()

            if not next:
                # if there are no next tasks, we can terminate the graph
                outputs[task.name] = state.get_out()

            # push next tasks to the channel only if
            # the current task is not a termination
            for next_task_name in next:
                # we set the inputs of the next tasks
                # to the outputs of the current task
                next_task = tasks[next_task_name]

                # in majority of cases there will be only one handle name
                # however we need to handle the case when single output
                # is mapped to multiple inputs on the next node
                handle_names = []
                for k, v in next_task.handles_mapping:
                    if v == task.name:
                        handle_names.append(k)

                for handle_name in handle_names:
                    next_state = next_task.input_states[handle_name]
                    next_state.set_state_and_permits(state, 1)

                # push next tasks to the channel only if the task is not active
                # and current task is not a termination
                if not (next_task_name in active_tasks) and not is_termination:
                    active_tasks.add(next_task_name)
                    queue.put(
                        ScheduledTask(
                            status="Task",
                            task_name=next_task_name,
                        )
                    )

            # increment depth of the finished task
            depths[task_id] = depth + 1
        except NodeRunError as e:
            logger.exception(f"Execution failed [id: {task_id}]")

            error = Message(
                id=uuid.uuid4(),
                value=str(e),
                start_time=start_time,
                end_time=datetime.datetime.now(),
            )

            outputs[task.name] = error

            # terminate entire graph by sending err task
            queue.put(
                ScheduledTask(
                    status="Err",
                    task_name=None,
                )
            )

        except Exception:
            logger.exception(f"Execution failed [id: {task_id}]")
            error = Message(
                id=uuid.uuid4(),
                value="Unexpected server error",
                start_time=start_time,
                end_time=datetime.datetime.now(),
            )
            outputs[task.name] = error
            queue.put(
                ScheduledTask(
                    status="Err",
                    task_name=None,
                )
            )

    def execute_task(
        self,
        task: Task,
        queue: queue.Queue,
    ):
        return self.thread_pool_executor.submit(
            self.execute_task_inner,
            task,
            queue,
        )



from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
import logging
from typing import Optional, Union

from lmnr.types import ChatMessage
from lmnr_engine.engine import Engine
{% set function_names = cookiecutter._tasks.values() | selectattr('node_type', '!=', 'Input') | map(attribute='function_name') | join(', ') %}
from .nodes.functions import {{ function_names }}
from lmnr_engine.engine.task import Task


logger = logging.getLogger(__name__)


class PipelineRunnerError(Exception):
    pass


@dataclass
class PipelineRunOutput:
    value: Union[str, list[ChatMessage]]


# This class is not imported in other files and can be renamed to desired name
class {{ cookiecutter.class_name }}:
    thread_pool_executor: ThreadPoolExecutor

    def __init__(
        self, thread_pool_executor: Optional[ThreadPoolExecutor] = None
    ) -> None:
        # Set max workers to hard-coded value for now
        self.thread_pool_executor = (
            ThreadPoolExecutor(max_workers=10)
            if thread_pool_executor is None
            else thread_pool_executor
        )

    def run(
        self,
        inputs: dict[str, Union[str, list]],
        env: dict[str, str] = {},
    ) -> dict[str, PipelineRunOutput]:
        """
        Run the pipeline with the given graph

        Raises:
            PipelineRunnerError: if there is an error running the pipeline
        """
        logger.info("Running pipeline {{ cookiecutter.pipeline_name }}, pipeline_version: {{ cookiecutter.pipeline_version_name }}")

        run_inputs = {}
        for inp_name, inp in inputs.items():
            if isinstance(inp, str):
                run_inputs[inp_name] = inp
            else:
                assert isinstance(inp, list), f"Invalid input type: {type(inp)}"
                run_inputs[inp_name] = [ChatMessage.model_validate(msg) for msg in inp]

        tasks = []
        {% for task in cookiecutter._tasks.values() %}
        tasks.append(
            Task(
                name="{{ task.name }}",
                value={{ "''" if task.node_type == "Input" else task.function_name }},
                handles_mapping={{ task.handles_mapping }},
                prev=[
                    {% for prev in task.prev %}
                    "{{ prev }}",
                    {% endfor %}
                ],
                next=[
                    {% for next in task.next %}
                    "{{ next }}",
                    {% endfor %}
                ],
            )
        )
        {% endfor %}
        engine = Engine.with_tasks(tasks, self.thread_pool_executor, env=env)

        # TODO: Raise PipelineRunnerError with node_errors
        run_res = engine.run(run_inputs)
        return {
            name: PipelineRunOutput(value=output.value)
            for name, output in run_res.items()
        }
